/*
  sendrecv -- a simple program for piping data to and from spamdyke
  Copyright (C) 2007 Sam Clippinger (samc (at) silence (dot) org)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <errno.h>
#include <getopt.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/select.h>
#include <time.h>
#include <stdint.h>
#include <ctype.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include "base64.h"
#include "md5.h"

#define STRLEN(X)               (sizeof(X) - 1)

#define DEFAULT_TIMEOUT_SECS    10
#define DEFAULT_WRITE_DELAY_SECS        2

#define STDIN_FD                0
#define STDOUT_FD               1
#define LINE_TERMINATOR_FAUX    '\r'
#define LINE_TERMINATOR         '\n'
#define CONTINUATION_MARKER     '-'
#define CONTINUATION_POSITION   3
#define MAX_BUF                 4095
#define MINVAL(a,b)             (((a) < (b)) ? (a) : (b))
#define MAXVAL(a,b)             (((a) > (b)) ? (a) : (b))
#define SMTP_DATA               "DATA"
#define SMTP_DATA_END           "."
#define SMTP_STARTTLS           "STARTTLS"
#define IPAD_BYTE       0x36
#define OPAD_BYTE       0x5C
#define CHALLENGE_CODE          "334 "
#define STARTTLS_CODE           "220 "

#define TLS_STATE_NONE          0
#define TLS_STATE_REQUESTED     1
#define TLS_STATE_ACTIVE        2

#define INPUT_STDIN             0
#define INPUT_CHILD             1

void usage()
  {
  printf("USAGE: sendrecv [ -d INITIAL_DELAY ] [ -t TIMEOUT_SECS ] [ -r DESIRED_RESPONSE ] [ -u USERNAME -p PASSWORD ] [ -- ] COMMANDLINE [ ARG1 ARG2 ... ]\n");

  exit(0);

  return;
  }

int cram_md5(char *return_text, char *username, char *password, char *encoded_challenge)
  {
  unsigned char challenge[MAX_BUF + 1];
  int strlen_challenge;
  unsigned char ipad[MAX_BUF + 1];
  unsigned char opad[MAX_BUF + 1];
  unsigned char result[16];
  unsigned char prepend[MAX_BUF + 1];
  int strlen_prepend;
  unsigned char secret[64];
  int i;

  for (i = 0; i < 64; i++)
    {
    ipad[i] = IPAD_BYTE;
    opad[i] = OPAD_BYTE;
    }

  if (strlen(password) > 64)
    {
    md5(secret, (unsigned char *)password, strlen(password));
    for (i = 16; i < 64; i++)
      secret[i] = '\0';
    }
  else
    {
    strncpy((char *)secret, password, strlen(password));
    for (i = strlen(password); i < 64; i++)
      secret[i] = '\0';
    }

  for (i = 0; i < 64; i++)
    {
    ipad[i] ^= secret[i];
    opad[i] ^= secret[i];
    }

  strlen_challenge = base64_decode(challenge, MAX_BUF, (unsigned char *)encoded_challenge, strlen(encoded_challenge));

  for (i = 0; i < strlen_challenge; i++)
    ipad[i + 64] = challenge[i];

  md5(opad + 64, ipad, strlen_challenge + 64);
  md5(result, opad, 80);

  strlen_prepend = snprintf((char *)prepend, MAX_BUF, "%s ", username);
  for (i = 0; i < 16; i++)
    snprintf((char *)(prepend + strlen_prepend + (i * 2)), MAX_BUF, "%.2x", result[i]);
  strlen_prepend += 32;

  return(base64_encode((unsigned char *)return_text, MAX_BUF, prepend, strlen_prepend));
  }

void tls_error(SSL *tls_session, int return_code)
  {
  char error_text[MAX_BUF + 1];
  int strlen_error_text;
  int saved_errno;
  int tls_error;
  int get_error;

  saved_errno = errno;
  strlen_error_text = 0;

  get_error = SSL_get_error(tls_session, return_code);

  error_text[0] = '\0';
  while (((tls_error = ERR_get_error()) != 0) &&
         (strlen_error_text < MAX_BUF))
    {
    if (strlen_error_text > 0)
      strlen_error_text += snprintf(error_text + strlen_error_text, MAX_BUF - strlen_error_text, ", ");

    ERR_error_string_n(tls_error, error_text + strlen_error_text, MAX_BUF - strlen_error_text);
    strlen_error_text += strlen(error_text + strlen_error_text);
    }

  switch (get_error)
    {
    case SSL_ERROR_NONE:
      // No error occurred.
      printf("(No SSL error)\n");
      break;
    case SSL_ERROR_ZERO_RETURN:
      // SSL connection closed
      printf("(SSL connection closed: %s)\n", error_text);
      break;
    case SSL_ERROR_WANT_READ:
    case SSL_ERROR_WANT_WRITE:
      // operation did not complete, call it again
    case SSL_ERROR_WANT_CONNECT:
    case SSL_ERROR_WANT_ACCEPT:
      // operation did not complete, call it again
    case SSL_ERROR_WANT_X509_LOOKUP:
      // callback function wants another callback.  Call the SSL function again.
      printf("(SSL function wants a callback: %s)\n", error_text);
      break;
    case SSL_ERROR_SYSCALL:
      // check the SSL error queue.  If return_code == 0, EOF found.  If return_code == -1, check errno.
      if (return_code == 0)
        strlen_error_text += snprintf(error_text + strlen_error_text, MAX_BUF - strlen_error_text, ", %s", "TLS EOF found");
      else if (return_code == -1)
        strlen_error_text += snprintf(error_text + strlen_error_text, MAX_BUF - strlen_error_text, ", %s", strerror(saved_errno));

      printf("(SSL failed due to a syscall: %s)\n", error_text);

      break;
    case SSL_ERROR_SSL:
      // Library failure, check the SSL error queue.
      printf("(SSL library failed: %s)\n", error_text);
      break;
    default:
      printf("(SSL unknown error: %s)\n", error_text);
      break;
    }

  fflush(NULL);

  return;
  }

int main(int argc, char *argv[], char *envp[])
  {
  int i;
  int child_pid;
  int child_stdout_pipe[2];
  int child_stdin_pipe[2];
  int timeout_secs;
  int opt;
  int tmp_int;
  char *desired_response;
  int strlen_desired_response;
  fd_set read_fds;
  fd_set write_fds;
  struct timeval tmp_timeout;
  time_t last_input_time;
  time_t last_write_time;
  time_t start_time;
  int write_delay_secs;
  char stdout_buf[MAX_BUF + 1];
  int strlen_stdout_buf;
  char *child_stdin_ptr;
  char stdin_buf[MAX_BUF + 1];
  int strlen_stdin_buf;
  int last_input;
  int inside_data;
  char *username;
  char *password;
  char insert_buf[MAX_BUF + 1];
  int strlen_insert_buf;
  char *child_insert_ptr;
  int tls_state;
  SSL_CTX *tls_context;
  SSL *tls_session;
  int initial_delay_secs;
  int ssl_return;

  opterr = 0;

  timeout_secs = DEFAULT_TIMEOUT_SECS;
  write_delay_secs = DEFAULT_WRITE_DELAY_SECS;
  strlen_desired_response = 0;
  desired_response = NULL;
  username = NULL;
  password = NULL;
  tls_state = TLS_STATE_NONE;
  tls_context = NULL;
  tls_session = NULL;
  initial_delay_secs = 0;

  while ((opt = getopt(argc, argv, "d:p:r:t:u:w:")) != -1)
    {
    switch (opt)
      {
      case 'd':
        if ((sscanf(optarg, "%d", &tmp_int) == 1) &&
            (tmp_int >= 0))
          initial_delay_secs = tmp_int;

        break;
      case 'p':
        password = optarg;
        break;
      case 't':
        if ((sscanf(optarg, "%d", &tmp_int) == 1) &&
            (tmp_int >= 0))
          timeout_secs = tmp_int;

        break;
      case 'r':
        strlen_desired_response = ((desired_response = optarg) != NULL) ? strlen(desired_response) : 0;

        break;
      case 'u':
        username = optarg;
        break;
      case 'w':
        if ((sscanf(optarg, "%d", &tmp_int) == 1) &&
            (tmp_int >= 0))
          write_delay_secs = tmp_int;

        break;
      default:
        usage();
        break;
      }
    }

  if (argv[optind] == NULL)
    usage();

  if (!pipe(child_stdout_pipe) != -1)
    if (!pipe(child_stdin_pipe) != -1)
      if ((child_pid = fork()) > 0)
        {
        close(child_stdout_pipe[1]);
        close(child_stdin_pipe[0]);

        start_time = time(NULL);
        last_input_time = time(NULL);
        strlen_stdout_buf = 0;
        child_stdin_ptr = stdin_buf;
        strlen_stdin_buf = 0;
        last_write_time = 0;
        inside_data = 0;
        strlen_insert_buf = 0;
        child_insert_ptr = NULL;
        last_input = INPUT_STDIN;

        while ((tmp_timeout.tv_sec = ((initial_delay_secs + timeout_secs) - (time(NULL) - last_input_time))) > 0)
          {
          tmp_timeout.tv_usec = 0;

          FD_ZERO(&read_fds);
          FD_SET(STDIN_FD, &read_fds);
          FD_SET(child_stdout_pipe[0], &read_fds);

          FD_ZERO(&write_fds);
          FD_SET(child_stdin_pipe[1], &write_fds);

          if (select(MAXVAL(child_stdout_pipe[0], child_stdin_pipe[1]) + 1, &read_fds, &write_fds, NULL, &tmp_timeout) != -1)
            {
            if (FD_ISSET(STDIN_FD, &read_fds) &&
                (strlen_stdin_buf < MAX_BUF) &&
                (read(STDIN_FD, stdin_buf + strlen_stdin_buf, 1) != 0))
              strlen_stdin_buf++;

            if (((time(NULL) - start_time) >= initial_delay_secs) &&
                (last_write_time == 0))
              last_input = INPUT_CHILD;

            if (FD_ISSET(child_stdin_pipe[1], &write_fds) &&
                ((child_insert_ptr != NULL) ||
                 (child_stdin_ptr < (stdin_buf + strlen_stdin_buf))) &&
                ((time(NULL) - last_write_time) >= write_delay_secs) &&
                ((last_input == INPUT_CHILD) ||
                 inside_data))
              {
              if (child_insert_ptr != NULL)
                {
                if (tls_state == TLS_STATE_ACTIVE)
                  SSL_write(tls_session, child_insert_ptr, 1);
                else
                  {
                  write(child_stdin_pipe[1], child_insert_ptr, 1);
                  fflush(NULL);
                  }

                if (child_insert_ptr == (insert_buf + strlen_insert_buf - 1))
                  {
                  last_input = INPUT_STDIN;
                  last_write_time = time(NULL);

                  strlen_insert_buf = 0;
                  child_insert_ptr = NULL;
                  }
                else
                  child_insert_ptr++;
                }
              else
                {
                if (tls_state == TLS_STATE_ACTIVE)
                  SSL_write(tls_session, child_stdin_ptr, 1);
                else
                  {
                  write(child_stdin_pipe[1], child_stdin_ptr, 1);
                  fflush(NULL);
                  }

                child_stdin_ptr++;

                if ((child_stdin_ptr - 1)[0] == LINE_TERMINATOR)
                  {
                  last_input = INPUT_STDIN;
                  last_write_time = time(NULL);

                  if (!inside_data &&
                      ((child_stdin_ptr - stdin_buf) >= (STRLEN(SMTP_DATA) + 1)) &&
                      ((strncasecmp(child_stdin_ptr - (STRLEN(SMTP_DATA) + 1), SMTP_DATA, STRLEN(SMTP_DATA)) == 0) ||
                       (strncasecmp(child_stdin_ptr - (STRLEN(SMTP_DATA) + 2), SMTP_DATA, STRLEN(SMTP_DATA)) == 0) ||
                       (strncasecmp(child_stdin_ptr - (STRLEN(SMTP_DATA) + 3), SMTP_DATA, STRLEN(SMTP_DATA)) == 0)))
                    {
                    inside_data = 1;
                    last_input = INPUT_CHILD;
                    }
                  else if (inside_data &&
                           ((child_stdin_ptr - stdin_buf) >= (STRLEN(SMTP_DATA_END) + 1)) &&
                           ((strncasecmp(child_stdin_ptr - (STRLEN(SMTP_DATA_END) + 1), SMTP_DATA_END, STRLEN(SMTP_DATA_END)) == 0) ||
                            (strncasecmp(child_stdin_ptr - (STRLEN(SMTP_DATA_END) + 2), SMTP_DATA_END, STRLEN(SMTP_DATA_END)) == 0) ||
                            (strncasecmp(child_stdin_ptr - (STRLEN(SMTP_DATA_END) + 3), SMTP_DATA_END, STRLEN(SMTP_DATA_END)) == 0)))
                    inside_data = 0;
                  else if (!inside_data &&
                           (tls_state == TLS_STATE_NONE) &&
                           ((child_stdin_ptr - stdin_buf) >= (STRLEN(SMTP_STARTTLS) + 1)) &&
                           ((strncasecmp(child_stdin_ptr - (STRLEN(SMTP_STARTTLS) + 1), SMTP_STARTTLS, STRLEN(SMTP_STARTTLS)) == 0) ||
                            (strncasecmp(child_stdin_ptr - (STRLEN(SMTP_STARTTLS) + 2), SMTP_STARTTLS, STRLEN(SMTP_STARTTLS)) == 0) ||
                            (strncasecmp(child_stdin_ptr - (STRLEN(SMTP_STARTTLS) + 3), SMTP_STARTTLS, STRLEN(SMTP_STARTTLS)) == 0)))
                    tls_state = TLS_STATE_REQUESTED;

                  if (child_stdin_ptr == (stdin_buf + strlen_stdin_buf))
                    {
                    child_stdin_ptr = stdin_buf;
                    strlen_stdin_buf = 0;
                    }
                  }
                }
              }

            if (FD_ISSET(child_stdout_pipe[0], &read_fds) ||
                ((tls_state == TLS_STATE_ACTIVE) &&
                 (SSL_pending(tls_session) >= 1)))
              {
              if (((tls_state == TLS_STATE_ACTIVE) &&
                   (SSL_read(tls_session, stdout_buf + strlen_stdout_buf, 1) != 0)) ||
                  ((tls_state != TLS_STATE_ACTIVE) &&
                   (read(child_stdout_pipe[0], stdout_buf + strlen_stdout_buf, 1) != 0)))
                {
                if (((strlen_stdout_buf + 1) == MAX_BUF) ||
                    (stdout_buf[strlen_stdout_buf] == LINE_TERMINATOR))
                  {
                  printf("%.*s", strlen_stdout_buf + 1, stdout_buf);
                  fflush(NULL);

                  if ((strlen_desired_response > 0) &&
                      (strncasecmp(stdout_buf, desired_response, MINVAL(strlen_stdout_buf, strlen_desired_response)) == 0))
                    break;

                  if ((username != NULL) &&
                      (password != NULL) &&
                      (strlen_stdout_buf >= STRLEN(CHALLENGE_CODE)) &&
                      (strncasecmp(stdout_buf, CHALLENGE_CODE, STRLEN(CHALLENGE_CODE)) == 0))
                    {
                    for (i = strlen_stdout_buf; i >= 0; i--)
                      if (!isspace(stdout_buf[i]))
                        {
                        stdout_buf[i + 1] = '\0';
                        break;
                        }

                    strlen_insert_buf = cram_md5(insert_buf, username, password, stdout_buf + STRLEN(CHALLENGE_CODE));
                    insert_buf[strlen_insert_buf] = LINE_TERMINATOR_FAUX;
                    insert_buf[strlen_insert_buf + 1] = LINE_TERMINATOR;
                    insert_buf[strlen_insert_buf + 2] = '\0';
                    strlen_insert_buf += 2;

                    child_insert_ptr = insert_buf;
                    last_input = INPUT_CHILD;
                    }
                  else if ((tls_state == TLS_STATE_REQUESTED) &&
                           (strlen_stdout_buf >= STRLEN(STARTTLS_CODE)) &&
                           (strncasecmp(stdout_buf, STARTTLS_CODE, STRLEN(STARTTLS_CODE)) == 0))
                    {
                    printf("(Negotiating TLS session. This may take a moment.)\n");
                    fflush(NULL);

                    SSL_library_init();
                    if (((tls_context = SSL_CTX_new(SSLv23_client_method())) != NULL) &&
                        ((tls_session = SSL_new(tls_context)) != NULL) &&
                        SSL_set_rfd(tls_session, child_stdout_pipe[0]) &&
                        SSL_set_wfd(tls_session, child_stdin_pipe[1]) &&
                        ((ssl_return = SSL_connect(tls_session)) == 1))
                      {
                      printf("(TLS session started.)\n");
                      fflush(NULL);
                      tls_state = TLS_STATE_ACTIVE;
                      }
                    else
                      {
                      tls_error(tls_session, ssl_return);

                      if (tls_session != NULL)
                        {
                        SSL_free(tls_session);
                        tls_session = NULL;
                        }

                      if (tls_context != NULL)
                        {
                        SSL_CTX_free(tls_context);
                        tls_context = NULL;
                        }

                      tls_state = TLS_STATE_NONE;
                      }

                    last_input = INPUT_CHILD;
                    }
                  else if ((strlen_stdout_buf < (CONTINUATION_POSITION + 1)) ||
                           (stdout_buf[CONTINUATION_POSITION] != CONTINUATION_MARKER))
                    last_input = INPUT_CHILD;

                  strlen_stdout_buf = 0;
                  }
                else
                  strlen_stdout_buf++;

                last_input_time = time(NULL);
                }
              else
                break;
              }
            }
          else
            {
            printf("sendrecv error: unable to select: %s\n", strerror(errno));
            fflush(NULL);
            break;
            }
          }

        if (tls_state == TLS_STATE_ACTIVE)
          {
          SSL_shutdown(tls_session);

          if (tls_session != NULL)
            {
            SSL_free(tls_session);
            tls_session = NULL;
            }

          if (tls_context != NULL)
            {
            SSL_CTX_free(tls_context);
            tls_context = NULL;
            }

          printf("(TLS session closed.)\n");
          fflush(NULL);
          }

        close(child_stdout_pipe[0]);
        close(child_stdin_pipe[1]);
        }
      else if (child_pid == 0)
        {
        close(child_stdout_pipe[0]);
        close(child_stdin_pipe[1]);

        if ((dup2(child_stdout_pipe[1], STDOUT_FD) != -1) &&
            (dup2(child_stdin_pipe[0], STDIN_FD) != -1))
          {
          if (execve(argv[optind], argv + optind, envp) == -1)
            {
            printf("sendrecv error: unable to exec %s: %s\n", argv[optind], strerror(errno));
            fflush(NULL);
            close(child_stdout_pipe[1]);
            close(child_stdin_pipe[0]);
            }
          }
        else
          {
          printf("sendrecv error: unable to reassign file descriptor: %s\n", strerror(errno));
          fflush(NULL);
          close(child_stdout_pipe[1]);
          close(child_stdin_pipe[0]);
          }
        }
      else
        {
        printf("sendrecv error: unable to fork: %s\n", strerror(errno));
        fflush(NULL);

        close(child_stdout_pipe[0]);
        close(child_stdout_pipe[1]);
        close(child_stdin_pipe[0]);
        close(child_stdin_pipe[1]);
        }
    else
      {
      printf("sendrecv error: unable to open pipe: %s\n", strerror(errno));
      fflush(NULL);

      close(child_stdout_pipe[0]);
      close(child_stdout_pipe[1]);
      }
  else
    {
    printf("sendrecv error: unable to open pipe: %s\n", strerror(errno));
    fflush(NULL);
    }

  return(0);
  }
